##############################################################################
# Imports
##############################################################################

import os.path

import sys

import subprocess
import argparse
import shutil
from catkin_pkg.packages import find_packages

##############################################################################
# Local imports
##############################################################################

import common
import console
from make_doc_html_templates import html_header, html_footer

import catkin_make.terminal_color as terminal_color
from catkin_make.terminal_color import fmt

##############################################################################
# Methods
##############################################################################

DOC_PROGRAM = 'rosdoc_lite'


def generate_doc(name, pkg_path, doc_path):
    document_path = doc_path + '/' + name
    args = [DOC_PROGRAM, '-o', document_path, pkg_path]
    output = subprocess.call(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return output


def output(fd, html):
    for h in html:
        os.write(fd, h)
        os.write(fd, '\n')


def generates_index_page(doc_path, pkg_names):

    index_page = doc_path + '/index.html'
    fd = os.open(index_page, os.O_RDWR | os.O_CREAT)

    output(fd, html_header)

    for pkg in pkg_names:
        link_html = '  <p><a href="' + pkg + '/html/index.html">' + pkg + '</a></p>'
        output(fd, [link_html])

    output(fd, html_footer)
    os.close(fd)


def make_doc(source_path, doc_path, packages):

    if not common.which(DOC_PROGRAM):
        console.error('[' + DOC_PROGRAM + '] is not available.')
        console.error('Please make sure [' + DOC_PROGRAM + '] is in your python path')
        return

    if not os.path.exists(doc_path):
        os.mkdir(doc_path)

    # List up packages with its absolute path
    packages_by_name = {p.name: source_path + '/' + path for path, p in packages.iteritems()}

    doc_output = {}
    console.pretty_println('Generating documents in ' + doc_path, console.cyan)
    for name, path in packages_by_name.items():
        console.pretty_println('  ' + name)
        output = generate_doc(name, path, doc_path)
        doc_output[name] = output

    generates_index_page(doc_path, packages_by_name.keys())

    console.pretty_println('')
    console.pretty_println('Document generation result. 0 may mean error. But it is fine most of time', console.bold_white)
    for name, err in doc_output.items():
        console.pretty_print(name, console.cyan)
        console.pretty_print(' : ')
        console.pretty_println(str(err))



##############################################################################
# Imports
##############################################################################

import os
import sys
import multiprocessing

##############################################################################
# Local imports
##############################################################################

import console
import python_setup
import config_cache
import subprocess
import ast

##############################################################################
# Methods
##############################################################################


def which(program):
    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    fpath, unused_fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def is_same_dir(dir_a, dir_b):
    rel_path = os.path.relpath(os.path.abspath(dir_a), os.path.abspath(dir_b))
    if rel_path == ".":
        return True
    else:
        return False


#def modified_environment(catkin_python_path, env=None):
#    '''
#      Prepends the path to the PYTHONPATH environment variable.
#      Could use some checking to make sure it is a single path, not a
#      pathsep list.
#    '''
#    if not env:
#        env = os.environ.copy()
#    try:
#        env['PYTHONPATH'] = env['PYTHONPATH'] + os.pathsep + catkin_python_path
#    except KeyError:
#        env['PYTHONPATH'] = catkin_python_path

def good_number_of_jobs():
    if 'ROS_PARALLEL_JOBS' in os.environ:
        jobs = os.environ['ROS_PARALLEL_JOBS']
    else:
        try:
            jobs = multiprocessing.cpu_count()
        except NotImplementedError:
            jobs = 1
    return jobs


def override_filename():
    return os.path.join(os.path.dirname(__file__), 'cmake', 'overrides.cmake')


def parent_directory(path):
    return os.path.abspath(os.path.join(path, os.pardir))


def create_symlink(src, dst, quiet=False):
    """
    Creates a symlink at dst to src.
    """
    if not os.path.exists(src):
        raise RuntimeError("'%s' is not a valid path" % src)
    try:
        os.symlink(src, dst)
        if not quiet:
            console.pretty_print('Creating symlink', console.white)
            console.pretty_print(' "%s" ' % dst, console.bold)
            console.pretty_print("->", console.white)
            console.pretty_println(' "%s." ' % src, console.bold)
    except Exception as ex_symlink:
        raise RuntimeError("Could not symlink '%s' to %s [%s]." % (src, dst, str(ex_symlink)))


def is_broken_symlink(link_path):
    '''
      Checks to see if the provided link_path is firstly, a link, and secondly if
      that link is broken.
    '''
    if os.path.islink(link_path):
        target_path = os.readlink(link_path)
        # Resolve relative symlinks
        if not os.path.isabs(target_path):
            target_path = os.path.join(os.path.dirname(link_path), target_path)
        if not os.path.exists(target_path):
            return True
    return False


def find_catkin(base_path, underlays_list=None):
    '''
      Search the underlays looking for catkin's toplevel.cmake and python module.
    '''
    if underlays_list is None:
        underlays_list = config_cache.get_underlays_list_from_config_cmake(base_path)
    catkin_toplevel = None
    catkin_python_path = None
    catkin_cmake_path = None
    for underlay in underlays_list:
        if os.path.isfile(os.path.join(underlay, 'share', 'catkin', 'cmake', 'toplevel.cmake')):
            # it is in an underlay's install space
            catkin_cmake_path = os.path.join(underlay, 'share', 'catkin', 'cmake')
            catkin_toplevel = os.path.join(underlay, 'share', 'catkin', 'cmake', 'toplevel.cmake')
            if os.path.isfile(os.path.join(underlay, python_setup.get_global_python_destination(), 'catkin', 'builder.py')):
                catkin_python_path = os.path.join(underlay, python_setup.get_global_python_destination())
            break
        elif os.path.isfile(os.path.join(underlay, python_setup.get_global_python_destination(), 'catkin', '__init__.py')):
            # it's probably a devel space
            console.error('Error: catkin seems to be buried in a chained devel space - not yet supporting this')
            # catkin_python_path = os.path.join(underlay, python_setup.get_global_python_destination())
            break
    return catkin_toplevel, catkin_python_path, catkin_cmake_path


def generate_underlay_environment(base_path):
    '''
      Parse all the underlay environment scripts to setup an initial environment for us to work in. This
      should be easier than manually crafting everything. We do maintain some more complexity than
      catkin_make though as we support a pyramid of underlays, not a linear chain.

      UNUSED and UNTESTED!
    '''
    underlays_list = config_cache.get_underlays_list_from_config_cmake(base_path)
    env = os.environ.copy()
    for underlay in reversed(underlays_list):
        print("\nUnderlay: %s" % underlay)
        env_script = os.path.join(underlay, 'env.sh')
        # Should validate that the script exists
        python_code = 'import os; print(dict(os.environ))'
        print [env_script, sys.executable, '-c', python_code]
        output = subprocess.check_output([env_script, sys.executable, '-c', python_code])
        print output
        # Not really positive what's going on here - this is what is in catkin.environment_cache.generate_environment_scripot
        # I think it just safely decodes strings in the dictionary.
        if hasattr(sys.stdout, 'encoding') and sys.stdout.encoding:
            output = output.decode(sys.stdout.encoding)
        env_after = ast.literal_eval(output)
        # calculate added and modified environment variables
        modified = {}
        for key, value in env_after.items():
            if key not in env:
                env[key] = value
            elif env[key] != value:
                # Need to be careful here - we are dealing with paths that will need to be merged.
                modified[key] = [env[key], value]
        # Deal with modified keys
        for key in sorted(modified.keys()):
            (old_value, new_value) = modified[key]
            if(key.contains('PATH') or old_value.contains(os.pathsep) or new_value.contains(os.pathsep)):
                # Merge paths
                old_elements = os.pathsep.split(old_value)
                new_elements = os.pathsep.split(new_value)
                for element in new_elements:
                    if not old_elements.contains(element):
                        env[key] = env[key] + os.pathsep + element
            else:
                # it's a regular variable - take the latest value assigned to it (not we're sorting underlays in the correct (reverse) order of priority
                env[key] = new_value
    return env


def generate_pkg_config_path(base_path):
    '''
      Generate a list of pkg_config_paths from the underlays

      @param base_path : used to find the config.cmake which lists the underlays
      @return list of paths
    '''
    underlays_list = config_cache.get_underlays_list_from_config_cmake(base_path)
    pkg_config_path = []
    for underlay in underlays_list:
        if os.path.isdir(os.path.join(underlay, 'lib', 'pkgconfig')):
            pkg_config_path.append(os.path.join(underlay, 'lib', 'pkgconfig'))
    return pkg_config_path


def get_default_paths(isolated=False):
    suffix = "_isolated" if isolated else ""
    base_path = os.environ.get("YUJIN_MAKE_ROOT") or os.environ.get("YUJIN_WORKSPACE") or os.getcwd()  # Fallback if os.environ.get returns None
    build_path = os.path.join(base_path, 'build' + suffix)
    devel_path = os.path.join(base_path, 'devel' + suffix)
    source_path = os.path.join(base_path, 'src')
    return (base_path, build_path, devel_path, source_path)



##############################################################################
# Imports
##############################################################################

import os.path

import sys
#import stat  # file permissions
import argparse
from argparse import RawTextHelpFormatter
#import shutil

##############################################################################
# Local imports
##############################################################################

import console
#import common

##############################################################################
# Constants
##############################################################################

DEFAULT_TRACK = "indigo"
VALID_TRACKS = ["groovy", "hydro", "indigo"]
DEFAULT_ROSINSTALL_DATABASE = 'https://raw.github.com/yujinrobot/yujin_tools/master/rosinstalls'

##############################################################################
# Methods
##############################################################################


def yujin_tools_home():
    """
    Get directory location of '.yujin_tools' directory (aka Yujin Tools home).

    @param env: override os.environ dictionary
    @type  env: dict
    @return: path to use use for log file directory
    @rtype: str
    """
    home_dir = os.path.join(os.path.expanduser('~'), '.yujin_tools')
    if not os.path.exists(home_dir):
        os.makedirs(home_dir)
    return home_dir


def get_default_track():
    filename = os.path.join(yujin_tools_home(), "track")
    try:
        f = open(filename, 'r')
    except IOError:
        set_default_track()
        return DEFAULT_TRACK
    track = f.read()
    f.close()
    return track


def set_default_track(track=DEFAULT_TRACK):
    if track not in VALID_TRACKS:
        raise RuntimeError("The track '%s' is not a valid track. Choose from %s\n" % (track, VALID_TRACKS))
    filename = os.path.join(yujin_tools_home(), "track")
    f = open(filename, 'w+')
    try:
        f.write(track.encode('utf-8'))
    finally:
        f.close()
    return track


def get_rosinstall_database_uri():
    filename = os.path.join(yujin_tools_home(), "rosinstall_database")
    try:
        f = open(filename, 'r')
    except IOError:
        return set_rosinstall_database_uri()
    rosinstall_database = f.read()
    f.close()
    return rosinstall_database


def set_rosinstall_database_uri(rosinstall_database=DEFAULT_ROSINSTALL_DATABASE):
    '''
      Set a uri for your rosinstall database.
    '''
    # could actually check that it is a valid uri though.
    filename = os.path.join(yujin_tools_home(), "rosinstall_database")
    f = open(filename, 'w+')
    try:
        f.write(rosinstall_database.encode('utf-8'))
    finally:
        f.close()
    return rosinstall_database

##############################################################################
# Utility Settings Script Functionality
##############################################################################


def help_string():
    overview = '\nThis is a convenience script for configuring yujin tools settings.\n\n'
    instructions = " \
 - 'yujin_tools_settings --get-default-track' : return the currently configured default track.\n \
 - 'yujin_tools_settings --set-default-track hydro' : save this track as the default track in yujin_tools_home.\n \
 - 'yujin_tools_settings --get-rosinstall-database-uri' : return the currently configured rosinstall database uri.\n \
 - 'yujin_tools_settings --set-rosinstall-database-uri' : save this url as the default rosinstall database uri.\n \
 "
    return overview + instructions


def parse_arguments():
    parser = argparse.ArgumentParser(description=help_string(), formatter_class=RawTextHelpFormatter)
    parser.add_argument('--get-default-track', action='store_true', help='print the default track that is being followed to screen')
    parser.add_argument('--set-default-track', action='store', default=None, help='set a new default track to work from %s' % VALID_TRACKS)
    parser.add_argument('--get-rosinstall-database-uri', action='store_true', help='print the default rosinstall database uri')
    parser.add_argument('--set-rosinstall-database-uri', action='store', default=None, help='set a new default  rosinstall database uri')
    args = parser.parse_args()
    return args


def main():
    args = parse_arguments()
    if args.get_default_track:
        #console.pretty_print("\nDefault Track: ", console.cyan)
        #console.pretty_println("%s\n" % get_default_track(), console.yellow)
        print get_default_track()
        sys.exit(0)
    if args.set_default_track:
        console.pretty_print("\nNew Default Track: ", console.cyan)
        console.pretty_println("%s\n" % set_default_track(args.set_default_track), console.yellow)
        sys.exit(0)
    if args.get_rosinstall_database_uri:
        print get_rosinstall_database_uri()
        sys.exit(0)
    if args.set_rosinstall_database_uri:
        console.pretty_print("\nNew Rosisntall Database Uri: ", console.cyan)
        console.pretty_println("%s\n" % set_rosinstall_database_uri(args.set_rosinstall_database_uri), console.yellow)
        sys.exit(0)
    print("%s" % help_string())
